package gnu.rfb;

/**
 * <p><br><br><center><table border="1" width="80%"><hr>
 * <strong><a href="http://www.amherst.edu/~tliron/vncj">VNCj</a></strong>
 * <p>
 * Copyright (C) 2000-2002 by Tal Liron
 * <p>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 * <p>
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * <a href="http://www.gnu.org/copyleft/lesser.html">GNU Lesser General Public License</a>
 * for more details.
 * <p>
 * You should have received a copy of the <a href="http://www.gnu.org/copyleft/lesser.html">
 * GNU Lesser General Public License</a> along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * <hr></table></center>
 **/

import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.IOException;

/**
 * CoRRE - Compact RRE Encoding.
 * <p>
 * We have an RRE header structure giving the number of subrectangles following.
 * Finally the data follows in the form [<bgpixel><subrect><subrect>...]
 * where each <subrect> is [<pixel><rfbCoRRERectangle>]. This means that the
 * whole rectangle must be at most 255x255 pixels.
 */
public class CoRRE extends RRE {

	/**
	 * Constructor for encoding an array of pixels into CoRRE.
	 * 
	 * @param pixels
	 *            The array of pixels to encode.
	 * @param pixelFormat
	 *            The current PixelFormat.
	 * @param offsetX
	 *            The offset in the X direction.
	 * @param offsetY
	 *            The offset in the Y direction.
	 * @param scanline
	 *            The width of a single scanline.
	 * @param x
	 *            The x-coordinate.
	 * @param y
	 *            The y-coordinate.
	 * @param w
	 *            The width.
	 * @param h
	 *            The height.
	 */
	public CoRRE(int[] pixels, PixelFormat pixelFormat, int offsetX,
			int offsetY, int scanline, int x, int y, int w, int h) {
		super(pixels, pixelFormat, offsetX, offsetY, scanline, x, y, w, h);
	}

	/**
	 * Constructor for encoding subrects into CoRRE.
	 * 
	 * @param x
	 *            The x-coordinate.
	 * @param y
	 *            The y-coordinate.
	 * @param w
	 *            The width.
	 * @param h
	 *            The height.
	 * @param pixelFormat
	 *            The current PixelFormat.
	 * @param bgpixel
	 *            The background pixel color.
	 * @param subrects
	 *            An array of Sub-Rectangles.
	 */
	public CoRRE(int x, int y, int w, int h, PixelFormat pixelFormat,
			int bgpixel, SubRect[] subrects) {
		super(x, y, w, h, pixelFormat, bgpixel, subrects);
	}

	/**
	 * Constructor for decoding CoRRE.
	 * 
	 * @param x
	 *            The x-coordinate.
	 * @param y
	 *            The y-coordinate.
	 * @param w
	 *            The width.
	 * @param h
	 *            The height.
	 * @param pixelFormat
	 *            The current PixelFormat.
	 * @param input
	 *            The input to read from.
	 * @throws IOException
	 *             If unable to read from the given input.
	 * 
	 * @author Ken Breeman (kbreeman@wpi.edu)
	 * @author Eric Griffel (sniffels@wpi.edu)
	 */
	public CoRRE(int x, int y, int w, int h, PixelFormat pixelFormat,
			DataInputStream input) throws IOException {
		super(x, y, w, h, pixelFormat, input);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gnu.rfb.RRE#writeData(java.io.DataOutput)
	 */
	public void writeData(DataOutput output) throws IOException {
		output.writeShort(x);
		output.writeShort(y);
		output.writeShort(w);
		output.writeShort(h);
		output.writeInt(RFB.EncodingCoRRE);
		output.writeInt(subrects.length);
		writePixel(output, pixelFormat, bgpixel);
		for (int i = 0; i < subrects.length; i++) {
			writePixel(output, pixelFormat, subrects[i].pixel);
			output.writeByte(subrects[i].x);
			output.writeByte(subrects[i].y);
			output.writeByte(subrects[i].w);
			output.writeByte(subrects[i].h);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gnu.rfb.RRE#clone()
	 */
	public Object clone() throws CloneNotSupportedException {
		SubRect[] subrectsClone = new SubRect[subrects.length];
		for (int i = 0; i < subrects.length; i++) {
			subrectsClone[i] = new SubRect();
			subrectsClone[i].pixel = subrects[i].pixel;
			subrectsClone[i].x = subrects[i].x;
			subrectsClone[i].y = subrects[i].y;
			subrectsClone[i].w = subrects[i].w;
			subrectsClone[i].h = subrects[i].h;
		}

		return new CoRRE(x, y, w, h, pixelFormat, bgpixel, subrectsClone);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see gnu.rfb.RRE#decode()
	 * 
	 * @author Ken  Breeman (kbreeman@wpi.edu)
	 * @author Eric Griffel (sniffels@wpi.edu)
	 */
	public int[] decode() throws IOException {

		int numRects = input.readInt();
		System.out.println("numRects:" + numRects);

		this.bgpixel = Rect.readPixel(input, pixelFormat);

		// Init the array of pixels to the bgPixel color.
		int pixels[] = new int[w * h];
		for (int count = 0; count < pixels.length; count++) {
			pixels[count] = this.bgpixel;
		}

		int rectp = 0; // Pixel value for this rect.
		int rectx = 0; // X-position
		int recty = 0; // Y-position
		int rectw = 0; // Width
		int recth = 0; // Height

		// For each rect sent.
		for (int i = 0; i < numRects; i++) {
			// Read it in.
			rectp = Rect.readPixel(input, pixelFormat);
			rectx = (int) (input.readByte() & 0xFF);
			recty = (int) (input.readByte() & 0xFF);
			rectw = (int) (input.readByte() & 0xFF);
			recth = (int) (input.readByte() & 0xFF);

			// Apply it to the pixels array.
			for (int j = 0; j < (rectw * recth); j++) {
				// (recty*w + rectx) --> starts us off at the correct initial
				// offset
				// (j / rectw)*w --> gives us the Y component
				// (j % rectw) --> gives us the X component
				pixels[(recty * w + rectx) + (j / rectw) * w + (j % rectw)] = rectp;
			}
		}

		return pixels;
	}
}
