package gnu.rfb;

/**
 * <br><br><center><table border="1" width="80%"><hr>
 * <strong><a href="http://www.amherst.edu/~tliron/vncj">VNCj</a></strong>
 * <p>
 * Copyright (C) 2000-2002 by Tal Liron
 * <p>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 * <p>
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * <a href="http://www.gnu.org/copyleft/lesser.html">GNU Lesser General Public License</a>
 * for more details.
 * <p>
 * You should have received a copy of the <a href="http://www.gnu.org/copyleft/lesser.html">
 * GNU Lesser General Public License</a> along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * <hr></table></center>
 **/

import java.awt.color.ColorSpace;
import java.awt.image.DataBuffer;
import java.awt.image.DirectColorModel;
import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.io.PrintStream;

/**
 * Represents the pixel format of pixels being sent between client and server.
 */
public class PixelFormat {

	// The BGR233 Standard Pixel Format.
	public static final PixelFormat BGR233 = new PixelFormat(8, 8, false, true,
			7, 7, 3, 0, 3, 6);

	// The RGB888 Standard Pixel Format.
	public static final PixelFormat RGB888 = new PixelFormat(32, 24, false,
			true, 0xFF, 0xFF, 0xFF, 16, 8, 0);

	// The number of bits per pixel.
	public int bitsPerPixel; // 8, 16 or 32

	// The color depth of each pixel.
	public int depth; // 8 to 32

	// Whether or not the pixels are big-Endian.
	public boolean bigEndian;

	// Whether or not True Color should be used.
	public boolean trueColour; // False -> requires colour map

	public int redMax; // Ignored if trueColor = false
	public int greenMax; // Ignored if trueColor = false
	public int blueMax; // Ignored if trueColor = false
	public int redShift; // Ignored if trueColor = false
	public int greenShift; // Ignored if trueColor = false
	public int blueShift; // Ignored if trueColor = false

	private DirectColorModel directColorModel = null;
	private int redMask = -1, greenMask = -1, blueMask = -1;
	private int redFix = -1, greenFix = -1, blueFix = -1;

	/**
	 * Empty constructor for custom pixel formats.
	 */
	public PixelFormat() {
	}

	/**
	 * Default Constructor for a PixelFormat.
	 * 
	 * @param bitsPerPixel
	 * @param depth
	 * @param bigEndian
	 * @param trueColour
	 * @param redMax
	 * @param greenMax
	 * @param blueMax
	 * @param redShift
	 * @param greenShift
	 * @param blueShift
	 */
	public PixelFormat(int bitsPerPixel, int depth, boolean bigEndian,
			boolean trueColour, int redMax, int greenMax, int blueMax,
			int redShift, int greenShift, int blueShift) {
		this.bitsPerPixel = bitsPerPixel;
		this.depth = depth;
		this.bigEndian = bigEndian;
		this.trueColour = trueColour;
		this.redMax = redMax;
		this.greenMax = greenMax;
		this.blueMax = blueMax;
		this.redShift = redShift;
		this.greenShift = greenShift;
		this.blueShift = blueShift;
	}

	/**
	 * Reads in a PixelFormat data from input and creates an instance.
	 * 
	 * @param input
	 *            The input to read from.
	 * @throws IOException
	 *             If unable to read from the input.
	 */
	public PixelFormat(DataInput input) throws IOException {
		bitsPerPixel = input.readUnsignedByte();
		depth = input.readUnsignedByte();
		bigEndian = input.readUnsignedByte() == 1 ? true : false;
		trueColour = input.readUnsignedByte() == 1 ? true : false;
		redMax = input.readUnsignedShort();
		greenMax = input.readUnsignedShort();
		blueMax = input.readUnsignedShort();
		redShift = input.readUnsignedByte();
		greenShift = input.readUnsignedByte();
		blueShift = input.readUnsignedByte();
	}

	/**
	 * Creates a PixelFormat based on an existing one.
	 * 
	 * @param pixelFormat
	 *            The PixelFormat to copy from.
	 */
	public PixelFormat(PixelFormat pixelFormat) {
		this.bitsPerPixel = pixelFormat.bitsPerPixel;
		this.depth = pixelFormat.depth;
		this.bigEndian = pixelFormat.bigEndian;
		this.trueColour = pixelFormat.trueColour;
		this.redMax = pixelFormat.redMax;
		this.greenMax = pixelFormat.greenMax;
		this.blueMax = pixelFormat.blueMax;
		this.redShift = pixelFormat.redShift;
		this.greenShift = pixelFormat.greenShift;
		this.blueShift = pixelFormat.blueShift;
	}

	/**
	 * Writes the PixelFormat data to the output.
	 * 
	 * @param output
	 *            The output to write to.
	 * @throws IOException
	 *             If unable to write to the output.
	 */
	public void writeData(DataOutput output) throws IOException {
		output.writeByte(bitsPerPixel);
		output.writeByte(depth);
		output.writeByte(bigEndian ? 1 : 0);
		output.writeByte(trueColour ? 1 : 0);
		output.writeShort(redMax);
		output.writeShort(greenMax);
		output.writeShort(blueMax);
		output.writeByte(redShift);
		output.writeByte(greenShift);
		output.writeByte(blueShift);
	}

	/**
	 * Prints the PixelFormat data to a PrintStream.
	 * 
	 * @param stream
	 *            The stream to print to.
	 */
	public void print(PrintStream stream) {
		stream.println("Bits-per-pixel: " + bitsPerPixel);
		stream.println("Depth:          " + depth);
		stream.println("Big Endian:     " + bigEndian);
		stream.println("True Colour:    " + trueColour);
		if (trueColour) {
			stream.println("R max:   " + redMax);
			stream.println("G max:   " + greenMax);
			stream.println("B max:   " + blueMax);
			stream.println("R shift: " + redShift);
			stream.println("G shift: " + greenShift);
			stream.println("B shift: " + blueShift);
		}
	}

	/**
	 * Translates a pixel based on the format.
	 * 
	 * @param pixel
	 *            The pixel to translate.
	 * @return The translated version of the pixel.
	 */
	public int translatePixel(int pixel) {
		if (redFix != -1) {
			return ((pixel & redMask) >> redFix << redShift)
					| ((pixel & greenMask) >> greenFix << greenShift)
					| ((pixel & blueMask) >> blueFix << blueShift);
		} else {
			return pixel;
		}
	}

	/**
	 * UNTESTED
	 * 
	 * An attempt to de-translate a pixel based on format.
	 * 
	 * @param pixel
	 *            The pixel to de-translate.
	 * @return The original pixel value.
	 * 
	 * @author Ken Breeman (kbreeman@wpi.edu)
	 * @author Eric Griffel (sniffels@wpi.edu)
	 * 
	 * UNTESTED
	 */
	public int deTranslatePixel(int pixel) {
		if (redFix != -1) {
			return ((pixel & redMask) << redFix >> redShift)
					| ((pixel & greenMask) << greenFix >> greenShift)
					| ((pixel & blueMask) << blueFix >> blueShift);
		} else {
			return pixel;
		}
	}

	/**
	 * @return The DirectColorModel for this format.
	 */
	public DirectColorModel toDirectColorModel() {
		return new DirectColorModel(ColorSpace.getInstance(ColorSpace.CS_sRGB),
				depth, redMax << redShift, greenMax << greenShift,
				blueMax << blueShift, 0, true, DataBuffer.TYPE_INT);
	}

	/**
	 * @return The DirectColorModel for this format.
	 */
	public DirectColorModel getDirectColorModel() {
		return directColorModel;
	}

	/**
	 * 
	 * @param directColorModel
	 */
	public void setDirectColorModel(DirectColorModel directColorModel) {
		this.directColorModel = directColorModel;

		redMask = directColorModel.getRedMask();
		greenMask = directColorModel.getGreenMask();
		blueMask = directColorModel.getBlueMask();

		redFix = fixColorModel(0xFF, redMax, redMask);
		greenFix = fixColorModel(0xFF, greenMax, greenMask);
		blueFix = fixColorModel(0xFF, blueMax, blueMask);
	}

	/**
	 * 
	 * 
	 * @param max1
	 * @param max2
	 * @param mask
	 * @return
	 */
	private static int fixColorModel(int max1, int max2, int mask) {
		int fix = 0;
		for (; fix < 8; fix++) {
			if (max1 == max2)
				break;

			max1 >>= 1;
		}

		while ((mask & 1) == 0) {
			fix++;
			mask >>= 1;
		}

		return fix;
	}
}
