/**
 * Copyright and disclaimer and yea...
 * 
 * @author Ken Breeman  (kbreeman@wpi.edu)
 * @author Eric Griffel (sniffels@wpi.edu)
 */

package gnu.vnc;

import gnu.rfb.PixelFormat;
import gnu.rfb.RFB;
import gnu.rfb.client.FrameBufferUpdate;
import gnu.rfb.server.DesCipher;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Vector;

/**
 * Abstracts out the details of a VNC client. Once constructed it connects to
 * the given server and maintains a VNC connection. It spawns its own thread to
 * handle the details of updating and protocol details.
 * 
 * Key events can be sent. Mouse movements and clicks can be sent. Clipboard can
 * be get/set. A screenshot of the remote frame buffer can be retrieved.
 * 
 * @author Ken Breeman (kbreeman@wpi.edu)
 * @author Eric Griffel (sniffels@wpi.edu)
 * 
 */
public class VNCViewer implements Runnable {

	private String hostname;
	private int port;
	private String password;

	private Socket socket;
	private DataInputStream input;
	private DataOutputStream output;
	private boolean doRun; // Flag for whether or not to continue running the
							// thread.
	private Thread viewerThread;

	private float protocol;

	private int width;
	private int height;
	private PixelFormat currentPixelFormat;
	private String desktopName;
	private byte sharedFlag;
	private String clipboard;

	// private ImageData imgData;
	private int[] image;
	private Vector<byte[]> inputQueue;

	/**
	 * Creates a new VNCViewer which connects and authenticates.
	 * 
	 * @param s
	 *            The host, host:display, or host::port to connect to.
	 * @param p
	 *            The password to send.
	 * @throws Exception
	 */
	public VNCViewer(String s, String p) throws Exception {

		this.password = p;

		if (s.contains("::")) { // For hostname::port
			hostname = s.substring(0, s.indexOf(':'));
			port = Integer
					.parseInt(s.substring(s.indexOf(':') + 2, s.length()));
		} else if (s.contains(":") && !s.contains("::")) { // For hostname:port
			hostname = s.substring(0, s.indexOf(':'));
			port = 5900 + Integer.parseInt(s.substring(s.indexOf(':') + 1, s
					.length()));
		} else { // For hostname
			hostname = s.toString();
			//System.out.println("Assuming desktop 0.");
			port = 5900; // Default to :0
		}

		init();
	}

	/**
	 * Creates a new VNCViewer which connects and authenticates.
	 * 
	 * @param host -
	 *            The host name to connect to.
	 * @param display -
	 *            The display number.
	 * @throws Exception If unable to connect or comminicate or authenticate.
	 */
	public VNCViewer(String host, int display, String p) throws Exception {
		hostname = host;
		port = display + 5900;
		this.password = p;

		init();
	}

	/**
	 * Initializes the connection and performs the RFB handshake.
	 * @throws Exception If unable to connect or comminicate or authenticate.
	 */
	private void init() throws Exception {
		
		connect();

		handshake();

		// Default to shared, refactor this later...
		sharedFlag = 1;

		this.image = new int[width * height];

		int pixels[] = new int[width * height];
		for (int i = 0; i < width * height; i++) {
			pixels[i] = 0;
		}

		// start off with a completely blank screen.
		for (int i = 0; i < this.image.length; i++) {
			this.image[i] = 0;
		}

		// and an empty input queue
		this.inputQueue = new Vector<byte[]>();

		// self spawning thread
		this.viewerThread = new Thread(this);
		this.viewerThread.start();
	}

	/**
	 * Connects the viewer to the host.
	 * 
	 * @return boolean - Whether or not the connection was successful.
	 * @throws Exception 
	 */
	private void connect() throws Exception {
		// Check if the hostname and port are sane.
		if (hostname == null || hostname.equalsIgnoreCase("") || port < 1
				|| port > 65535) {
			throw new Exception("ERROR: Invalid host name or port number.");
		}

		// Connect socket.
		try {

			socket = new Socket(hostname, port);
			input = new DataInputStream(new BufferedInputStream(socket
					.getInputStream()));
			output = new DataOutputStream(new BufferedOutputStream(socket
					.getOutputStream(), 16384));

		} catch (UnknownHostException e) {
			//System.out.println("Unable to find host: " + hostname + "\n");
			throw new Exception("Unable to find the host.");
		} catch (IOException e) {
			//System.out.println("Unable to connect to host: " + hostname + "\n");
			throw new Exception("Unable to connect to the host.");
		}
	}

	/**
	 * Performs the RFB handshake with the host.
	 * 
	 * @return Whether or not the handshake was successful.
	 * @throws Exception 
	 */
	private void handshake() throws Exception {

		try {
			String protocolVersion = input.readLine();

			if (!protocolVersion.startsWith("RFB")) {
				
				//System.out.println("ERROR: Server at host is not an RFB server.");
				throw new Exception("ERRROR: The host you are trying to connect to is not a VNC host.");
			}

			int major = Integer.parseInt(protocolVersion.substring(4, 7));
			int minor = Integer.parseInt(protocolVersion.substring(8));

			protocol = (float) major + (float) ((float) minor / (float) 10.00);
			//System.out.println("host protocol:" + major + "." + minor + " ");
			//System.out.println("host protocol:" + protocol);

			if (major < 3 || (major == 3 && minor < 3)) {
				//System.out.println("ERROR: Server at host is using an older RFB protocol version than 3.3");
				socket.close();
				throw new Exception("ERROR: The host you are trying to connect to is using an older RFB protocol version than 3.3");
			}

			// The server supports at least 3.3 and so do we.
			// output.writeBytes("RFB 003.003\n");
			output.writeBytes(RFB.ProtocolVersionMsg);
			// output.writeBytes("RFB 003.008\n");
			output.flush();

			// /////////////////////////////////////////////////////////////////////
			// Now to negotiate security.
			if (major > 3 || (major == 3 && minor >= 7)) {

				int numberOfSecurityTypes = (int) input.readByte();
				// System.out.println(numberOfSecurityTypes);

				if (numberOfSecurityTypes == 0) {
					//System.out.println("ERROR: Zero security types available.");
					socket.close();
					throw new Exception("The server does not have any available authentication methods.");
				}

				if (numberOfSecurityTypes < 0 || numberOfSecurityTypes > 32) {
					//System.out.println("ERROR: Too many or too few security types listed by server.");
					socket.close();
					throw new Exception("The server does not have any available authentication methods.");
				}

				int securityTypes[] = new int[numberOfSecurityTypes];

				for (int i = 0; i < numberOfSecurityTypes; i++) {
					securityTypes[i] = (int) input.readByte();
					// System.out.println(securityTypes[i]);
				}
			}

			else if (major == 3 && minor == 3) {

				int securityType = input.readInt();

				if (securityType == RFB.ConnFailed) { // s == 0
					//System.out.println("Security type negotiation failed.");
					socket.close();
					throw new Exception("ERROR: Security type negotionation failed.");
				} else if (securityType == RFB.NoAuth) { // s == 1
					// type == none
				} else if (securityType == RFB.VncAuth) { // s == 2
					// type == vnc
					//System.out.println("INFO: Security Type VNC.");
					byte[] challengeBytes = new byte[16];

					input.readFully(challengeBytes);

					// Copied from DefaultRFBAuthenticator.java

					byte[] passBytes = new byte[8];
					for (int i = 0; i < passBytes.length; i++) {
						if (i < password.length()) {
							passBytes[i] = password.getBytes()[i];
						} else {
							passBytes[i] = 0;
						}
					}
					DesCipher cipher = new DesCipher(passBytes);

					byte[] encryptedBytes = new byte[16];

					cipher.encrypt(challengeBytes, 0, encryptedBytes, 0);
					cipher.encrypt(challengeBytes, 8, encryptedBytes, 8);

					output.write(encryptedBytes);
					output.flush();

					// System.out.println("Encrypted Password sent.");

					if (input.readInt() != RFB.VncAuthOK) { // == 0
						//System.out.println("ERROR: Invalid password, authentication failed.");
						socket.close();
						throw new Exception("ERROR: Invalid password, authentication failed.");
					}

				} else {
					//System.out.println("Security type negotiation failed.");
					socket.close();
					throw new Exception("Security type negotiation failed.");
				}
			}

			output.write(sharedFlag);
			output.flush();

			// Read in the server's init message.
			width = input.readShort();
			height = input.readShort();

			// System.out.println("width:" + width);
			// System.out.println("height:" + height);

			currentPixelFormat = new PixelFormat(input);

			//currentPixelFormat.print(System.out);

			// Padding!
			input.readByte();
			input.readByte();
			input.readByte();

			int nameLength = input.readInt();

			// System.out.println("nameLength:" + nameLength);
			byte[] name = new byte[nameLength];
			input.readFully(name, 0, nameLength);

			desktopName = new String(name);

			// System.out.println("name:" + desktopName);

		} catch (IOException e) {
			//System.out.println("ERROR: Handshake failed with IOException.");
			throw new Exception("ERROR: Unexpectadly disconnected from server.");
		}
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	public void run() {

		try {

			// Tell the server which encodings we support.
			// TODO: clean this up!

			output.writeByte(RFB.SetEncodings);
			output.writeByte((byte) 0); // padding
			output.writeShort(5); // number of encodings
			// encodings in order of preference
			output.writeInt(RFB.EncodingHextile);
			output.writeInt(RFB.EncodingCopyRect);
			output.writeInt(RFB.EncodingCoRRE);
			output.writeInt(RFB.EncodingRRE);
			output.writeInt(RFB.EncodingRaw);
			output.flush();

			// Tell the server which PixelFormat we prefer.
			output.write(RFB.SetPixelFormat);
			output.writeByte((byte) 0); // padding
			output.writeByte((byte) 0); // padding
			output.writeByte((byte) 0); // padding
			currentPixelFormat.writeData(output);
			output.writeByte((byte) 0); // padding
			output.writeByte((byte) 0); // padding
			output.writeByte((byte) 0); // padding
			output.flush();

			// Tell server we want an initial update.
			output.writeByte(RFB.FrameBufferUpdateRequest);
			output.writeByte((byte) 0); // padding
			output.writeShort(0); // x
			output.writeShort(0); // y
			output.writeShort(this.width); // w
			output.writeShort(this.height); // h
			output.flush();

		} catch (IOException e1) {
			//e1.printStackTrace();
			//System.err.println("ERROR: I/O error while beginning transactions with VNC server.");
			this.doRun = false;
			return;
		}

		doRun = true;

		// Process incoming messages and send outgoing messages from the queue.
		while (doRun) {

			// Send off messages that need to be sent.
			while (!this.inputQueue.isEmpty()) {
				try {
					output.write(this.popEvent());
					output.flush();
				} catch (IOException e) {
					System.err
							.println("ERROR: Client disconnected unexpectedly, unable to send event.");
					this.stop();
					break;
				}
			}

			try {
				if (input.available() < 1) {
					Thread.sleep(10);
				} else {
					// Read in the message type and then delegate accordingly.

					byte type = input.readByte();

					switch (type) {

					case RFB.FrameBufferUpdate:

						input.readByte(); // Padding

						int numberOfRectangles = input.readUnsignedShort();

						for (int i = 0; i < numberOfRectangles; i++) {

							FrameBufferUpdate update;
							try {
								update = new FrameBufferUpdate(
										currentPixelFormat, input, image,
										this.width, this.height);

							} catch (Exception e) {
								e.printStackTrace();
								this.doRun = false;
								System.err
										.println("ERROR: Encountered an exception while updating the frame buffer.");
								return;
							}

							this.setImage(update.getImage());

						}

						break;

					case RFB.SetColourMapEntries:

						// TODO: Support color maps.

						input.readByte(); // Padding
						int firstColor = input.readShort();
						int numberOfColors = input.readShort();

						
						for (int i = 0; i < numberOfColors; i++) {
							int r = input.readShort();
							int b = input.readShort();
							int g = input.readShort();

						}

					case RFB.Bell:

						// Toolkit.getDefaultToolkit().beep();

						// System.out.print("\007");
						// System.out.flush();

						// Play audio file.

					case RFB.ServerCutText:

						input.readByte(); // Padding

						int length = input.readInt();

						byte cutTextBytes[] = new byte[length];

						input.read(cutTextBytes, 0, length);

						String cutTextString = new String(cutTextBytes);

						this.setClipboard(cutTextString);

						break;
					}

				}
			} catch (IOException e) {
				System.out
						.println("ERROR: Client IO error, closing Viewer thread.");
				// e.printStackTrace();
				doRun = false;
				return;
			} catch (InterruptedException e) {
			}
		}

		// Cleanup

	}

	/**
	 * Stops the client thread gracefully.
	 */
	public void stop() {
		doRun = false;
	}

	/**
	 * Tests out the VNC viewer with some default settings.
	 * 
	 * @param args
	 */
	public static void main(String args[]) {

		VNCViewer client = null;
		try {
			client = new VNCViewer("localhost:50", "password");
		} catch (Exception e1) {
			e1.printStackTrace();
		}

		while (client.doRun) {
			try {
				Thread.sleep(6000000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}

		}
	}

	/**
	 * @return The clipboard of the remote host as a String.
	 */
	public String getClipboard() {
		return clipboard.toString();
	}

	/**
	 * @param clipboard
	 *            The clipboard string to set on the remote host.
	 */
	public void setClipboard(String clipboard) {
		this.clipboard = clipboard;
	}

	/**
	 * @return The name of the desktop on the remote host.
	 */
	public String getName() {
		return this.desktopName.toString();
	}

	/**
	 * @return The hostname::port of the connected server.
	 */
	public String getServerHost() {
		return (this.hostname + "::" + port + "");
	}

	/**
	 * @return The width of the VNC session.
	 */
	public int getWidth() {
		return this.width;
	}

	/**
	 * @return The height of the VNC session.
	 */
	public int getHeight() {
		return this.height;
	}

	/**
	 * @return The color depth of the VNC session.
	 */
	public int getDepth() {
		return this.currentPixelFormat.bitsPerPixel;
	}

	/**
	 * @return The current PixelFormat for the connection.
	 */
	public PixelFormat getPixelFormat() {
		return this.currentPixelFormat;
	}

	/**
	 * @return The image data of the remote desktop via RFB.
	 */
	public int[] getImage() {
		return this.image;
	}

	/**
	 * Sets the current image of the remote system's desktop.
	 * @param img The image of the remote system's desktop.
	 */
	private synchronized void setImage(int[] img) {
		this.image = img;
	}

	/**
	 * Pushes a client to server event as an array of bytes.
	 * 
	 * @param e The message to push already byte-encoded.
	 */
	private synchronized void pushEvent(byte[] e) {
		this.inputQueue.add(e);
	}

	/**
	 * Pops a client to server event as an array of bytes.
	 * 
	 * @param e The message as a byte-encoded array.
	 */
	private synchronized byte[] popEvent() {

		if (this.inputQueue.isEmpty()) {
			return null;
		} // else

		byte[] temp = this.inputQueue.firstElement();
		this.inputQueue.remove(0);
		return temp;
	}

	/**
	 * Sends a KeyEvent to the server.
	 * 
	 * @param key Which key
	 * @param down Whether or not the key is down (pressed) or up (released).
	 */
	public void sendKey(int key, boolean down) {
		// add a key event to the send queue

		byte[] keyEvent = new byte[8];
		keyEvent[0] = 4;
		if (down) {
			keyEvent[1] = 1;
		} else {
			keyEvent[1] = 0;
		}
		keyEvent[2] = 0;
		keyEvent[3] = 0;

		keyEvent[4] = (byte) ((key >> 24) & 0x00000FF);
		keyEvent[5] = (byte) ((key >> 16) & 0x00000FF);
		keyEvent[6] = (byte) ((key >> 8) & 0x00000FF);
		keyEvent[7] = (byte) ((key) & 0x000000FF);

		this.pushEvent(keyEvent);
	}

	/**
	 * Sends a mouse pointer event to the server.
	 * 
	 * @param x The x-coordinate of the event.
	 * @param y The y-coordinate of the event.
	 * @param buttons A bitmask of which buttons are pressed.
	 */
	public void sendMouse(int x, int y, byte buttons) {
		// add a mouse event to the send queue

		byte[] mouseEvent = new byte[6];
		mouseEvent[0] = 5;
		mouseEvent[1] = buttons;
		mouseEvent[2] = (byte) ((x >> 8) & 0x000000FF);
		mouseEvent[3] = (byte) ((x) & 0x000000FF);
		mouseEvent[4] = (byte) ((y >> 8) & 0x000000FF);
		mouseEvent[5] = (byte) ((y) & 0x000000FF);

		this.pushEvent(mouseEvent);
	}

	/**
	 * Requests a refresh of an area of the screen.
	 * 
	 * @param incremental Whether or not this is an incremental request.
	 * @param x The x-coordinate.
	 * @param y The y-coordinate.
	 * @param w The width.
	 * @param h The height.
	 */
	public void refresh(boolean incremental, int x, int y, int w, int h) {
		// add a fbUpdate event to the send queue

		byte[] refreshEvent = new byte[10];

		refreshEvent[0] = 3;
		if (incremental) {
			refreshEvent[1] = 1;
		} else {
			refreshEvent[1] = 0;
		}
		refreshEvent[2] = (byte) ((x >> 8) & 0x000000FF);
		refreshEvent[3] = (byte) (x & 0x000000FF);
		refreshEvent[4] = (byte) ((y >> 8) & 0x000000FF);
		refreshEvent[5] = (byte) (y & 0x000000FF);
		refreshEvent[6] = (byte) ((w >> 8) & 0x000000FF);
		refreshEvent[7] = (byte) (w & 0x000000FF);
		refreshEvent[8] = (byte) ((h >> 8) & 0x000000FF);
		refreshEvent[9] = (byte) (h & 0x000000FF);

		this.pushEvent(refreshEvent);
	}

	/**
	 * @return the protocol
	 */
	public float getProtocol() {
		return protocol;
	}

}
